"""
graph_objs
==========

This package imports definitions for all of Plotly's graph objects. For more
information, run help(Obj) on any of the following objects defined here.

The reason for the package graph_objs and the module graph_objs is to provide
a clearer API for users.

"""
from __future__ import absolute_import

from plotly.graph_objs.graph_objs import *  # this is protected with __all__
